/*****************************************************************************
*
* Atmel Corporation
*
* File              : usi_logger_slave.c (usi_twi_slave.c)
* Compiler          : WINAVR
* Revision          : 
* Date              : 09.06.08
*							 Fehler beseitigt, der 0x3e im Datenstrom als TWI-STOP interpretiert 
*							 11.06.08
*                     das Konzept der Datenpufferung komplett gendert
* Support mail      : 
*
* Supported devices : All device with USI module can be used.
*                     The example is written for the ATmega169, ATtiny26 & ATtiny2313
*
* AppNote           : AVR312 - Using the USI module as a TWI slave
*
* Description       : Functions for USI_TWI_receiver and USI_TWI_transmitter.
*
* Changes for WinAVR by Frank Jonischkies, fjonisch@gmx.de, 11.08.2005
* angepasste/abgespeckte Version fr den Datenlogger
* Tabwidth			  : 3
*
****************************************************************************/

#include <avr/io.h>
#include <avr/interrupt.h>
#include "usi_logger_slave.h"

extern volatile uint8_t RxTail;				// volatile !			
extern volatile uint8_t RxHead;					
extern uint8_t TWI_control[BUFFER_SIZE];	// Array fr die Steuerzeichen START, ACK, NACK, STOP
extern uint8_t TWI_data[BUFFER_SIZE];		// Array fr die Daten														

//-----------------------------------------------------------------------------
// Initialise USI for TWI Slave mode
// Bentigt keine Adresse, da nur (mit-) gelesen wird
void usi_twi_slave_init(void)
{
PORT_USI |=	(1<<PORT_USI_SCL);					// Set SCL high
PORT_USI |=	(1<<PORT_USI_SDA);					// Set SDA high
DDR_USI	|=	(1<<PORT_USI_SCL);					// Set SCL as output
DDR_USI &= ~(1<<PORT_USI_SDA);					// Set SDA as input
USICR 	=	(1<<USISIE)|(0<<USIOIE)|			// Enable Start Condition Interrupt. Disable Overflow Interrupt.
				(1<<USIWM1)|(0<<USIWM0)| 			// Set USI in Two-wire mode. No USI Counter overflow prior
															// to first Start Condition (potentail failure)
				(1<<USICS1)|(0<<USICS0)|			// Shift Register Clock Source = External, positive edge
				(0<<USICLK)|(0<<USITC);
USISR = 0xF0;											// Clear all flags and reset overflow counter
}


//-----------------------------------------------------------------------------
//  Detects the USI_TWI Start Condition and intialises the USI
ISR(USI_START_vect)
{
uint8_t register tmp_Rx_Head;						// spart einige Speicheraktionen
tmp_Rx_Head = RxHead;								// volatile in Register zwischenspeichern
USISR |= (1<<USIPF);									// Zur Sicherheit Stop-Flag lschen
															// Set default starting conditions for new TWI package
DDR_USI &= ~(1<<PORT_USI_SDA);					// Set SDA as input

															// Index hochzhlen (nchster "Datensatz")
if ((++tmp_Rx_Head) >= BUFFER_SIZE) tmp_Rx_Head = 0;
RxHead = tmp_Rx_Head;								// neuen Wert von RxHead wieder im RAM ablegen

TWI_control[tmp_Rx_Head] = START;				// Symbol fr START in TWI_control ablegen

															// warten, bis SCL auf low gegangen ist
															// die Verknpfung muss logisch UND (&&) sein nicht bitweise (&)
while ( (PIN_USI & (1<<PORT_USI_SCL)) && !(USISR & (1<<USIPF)) );
															// Wait for SCL to go low to ensure the "Start Condition" has completed.
                                             // If a Stop condition arises then leave the interrupt to prevent waiting forever.																// Enable TWI Coutner overflow and interrupts
USICR=(1<<USISIE)|(1<<USIOIE)|					// Enable Overflow and Start Condition Interrupt.
															// (Keep StartCondInt to detect RESTART)
		(1<<USIWM1)|(1<<USIWM0)| 					// Set USI in Two-wire mode.
		(1<<USICS1)|(0<<USICS0)|(0<<USICLK)| 	// Shift Register Clock Source = External, positive edge
		(0<<USITC);
USISR=(1<<USISIF)|(1<<USIOIF)|(1<<USIPF)|		// Clear flags
		(1<<USIDC) |(0x0<<USICNT0); 				// Set USI to sample 8 bits i.e. count 16 external pin toggles.
}

//-----------------------------------------------------------------------------

ISR(USI_OVERFLOW_vect)
{
uint8_t register tmp_Rx_Head;						// spart einige Speicheraktionen
tmp_Rx_Head = RxHead;								// volatile in Register zwischenspeichern
 
TWI_data[tmp_Rx_Head] = USIDR;					// eingegangene Daten in TWI_data speichern
USIDR=0; 												// prepare to receive NACK

															// Index hochzhlen (nchster "Datensatz")
if ((++tmp_Rx_Head) >= BUFFER_SIZE) tmp_Rx_Head = 0;

RxHead = tmp_Rx_Head;								// neuen Wert von RxHead wieder im RAM ablegen

USISR=(1<<USISIF)|(1<<USIOIF)|(0<<USIPF)|		// Clear all flags, 
		(1<<USIDC) |(0xE<<USICNT0);				// Set USI to shift out 1 bit

while (!(USISR & ((1<<USISIF)|(1<<USIOIF) )));// bis zum Overflow bzw. START warten

if (USIDR)												// wenn SDA high (nicht auf Masse gezogen),
	{
	TWI_control[tmp_Rx_Head] = NACK;				// dann ist das ein NACK
	}
else														// wenn SDA low (auf Masse gezogen), 
	{
	TWI_control[tmp_Rx_Head] = ACK;				// dann ist das ein ACK
	}

USISR=(1<<USISIF)|(1<<USIOIF)|(1<<USIPF)|		// Alle flags lschen, die nachfolgend abgefragt werden !!!
		(1<<USIDC) |(0xE<<USICNT0); 				// Set USI to sample 1 bit
															// Wait until 1 bit sampled OR STOP- OR START- condition.
while ( !(USISR & ( (1<<USISIF)|(1<<USIOIF)|(1<<USIPF) )) );

if ( USISR & (1<<USIPF) )							// If STOP Cond
	{														// Set USI in TWI Start Condition Mode
	USICR=(1<<USISIE)|(0<<USIOIE)|				// Enable Start Condition Interrupt. Disable Overflow Interrupt.
			(1<<USIWM1)|(0<<USIWM0)|				// Set USI in Two-wire mode. No USI Counter overflow hold.
			(1<<USICS1)|(0<<USICS0)|(0<<USICLK)|// Shift Register Clock Source = External, positive edge
			(0<<USITC);
	USISR=(0<<USISIF)|(1<<USIOIF)|(1<<USIPF)|	// Clear all flags but Start Condition
			(1<<USIDC) |(0x0<<USICNT0);
															// das STOP muss in TWI_control gespeichert werden, 
															// da ist aber das vorausgegangene N/ACK gespeichert
															// darum das vorausgegangene N/ACK in TWI_data kopieren
	TWI_data[tmp_Rx_Head] = TWI_control[tmp_Rx_Head];
	TWI_control[tmp_Rx_Head] = STOP;				// das STOP in TWI_control speichern
	}

else
	{
	USISR=(0<<USISIF)|(1<<USIOIF)|(0<<USIPF)|	// Clear all flags, except Start Cond
			(1<<USIDC) |(0x2<<USICNT0);			// Set USI to sample last 7 bits
	}
}
